/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  nfc-converter.c
 * 
 *  Copyright (C) 2006-2008 Brian Pepple
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Author: Brian Pepple <bpepple@fedoraproject.org>
 * 
 */

#ifdef HAVE_CONFIG_H
 #include <config.h> /* for GETTEXT_PACKAGE */
#endif

#include "nfc-converter.h"

#include <string.h>

#include <gio/gio.h>
#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <libnautilus-extension/nautilus-file-info.h>
 
typedef struct _NfcConverterPrivate NfcConverterPrivate;

struct _NfcConverterPrivate {
	GList		*files;

	int		 flacs_converted;
	int		 flacs_total;
	gboolean	 cancelled;

	gchar		 quality[2];
	gchar		*new_path;

	GtkDialog	*dialog;
	GtkSpinButton	*quality_spinbutton;
	GtkFileChooser	*path_chooser;

	GtkWidget	*progress_dialog;
	GtkWidget	*progress_bar;
	GtkWidget	*progress_label;
};

#define NFC_CONVERTER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), NFC_TYPE_CONVERTER, NfcConverterPrivate))

G_DEFINE_TYPE (NfcConverter, nfc_converter, G_TYPE_OBJECT)

enum {
	PROP_FILES = 1,
};

typedef enum {
	SIGNAL_TYPE_EXAMPLE,
	LAST_SIGNAL
} NfcConverterSignalType;

static void run_op (NfcConverter *converter);

static void
nfc_converter_finalize (GObject *object)
{
	NfcConverter 	    *converter = NFC_CONVERTER (object);
	NfcConverterPrivate *priv = NFC_CONVERTER_GET_PRIVATE (converter);

	g_free (priv->new_path);

	G_OBJECT_CLASS (nfc_converter_parent_class)->finalize(object);
}

static void
nfc_converter_set_property (GObject      *object,
			    guint         property_id,
			    const GValue *value,
			    GParamSpec   *pspec)
{
	NfcConverter        *self = NFC_CONVERTER (object);
	NfcConverterPrivate *priv = NFC_CONVERTER_GET_PRIVATE (self);

	switch (property_id) {
	case PROP_FILES:
		priv->files = g_value_get_pointer (value);
		priv->flacs_total = g_list_length (priv->files);
		break;
	default:
		/* We don't have any other property... */
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}

static void
nfc_converter_get_property (GObject    *object,
			    guint       property_id,
			    GValue     *value,
			    GParamSpec *pspec)
{
	NfcConverter	    *self = NFC_CONVERTER (object);
	NfcConverterPrivate *priv = NFC_CONVERTER_GET_PRIVATE (self);

	switch (property_id) {
	case PROP_FILES:
		g_value_set_pointer (value, priv->files);
		break;
	default:
		/* We don't have any other property... */
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}

static void
nfc_converter_class_init (NfcConverterClass *klass)
{
	g_type_class_add_private (klass, sizeof (NfcConverterPrivate));

	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	GParamSpec   *files_param_spec;

	object_class->finalize = nfc_converter_finalize;
	object_class->set_property = nfc_converter_set_property;
	object_class->get_property = nfc_converter_get_property;

	files_param_spec = g_param_spec_pointer ("files",
						 "Files",
						 "Set selected files",
						 G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);

	g_object_class_install_property (object_class,
					 PROP_FILES,
					 files_param_spec);
}

static char *
nfc_converter_transform_uri (NfcConverter *converter, GFile *file)
{
	NfcConverter	    *self = NFC_CONVERTER (converter);
	NfcConverterPrivate *priv = NFC_CONVERTER_GET_PRIVATE (self);

	gchar 		    *old_basename;
	gchar		    *new_basename;
	gchar		    *extension;
	GString		    *result;

	/* Let's get the basename from the GFile. */
	old_basename = g_file_get_basename (file);

	/* 
	 * Now let's drop the .flac extension,
	 * and replace it with the .ogg extension
	 */
	extension = g_strdup (strrchr (old_basename, '.'));
	if (extension != NULL)
		old_basename[strlen (old_basename) - strlen (extension)] = '\0';
	g_free (extension);

	new_basename = g_strdup_printf ("%s%s", old_basename, ".ogg");
	g_free (old_basename);

	/*
	 * Ok, let's create a new uri from the save path,
	 * and the new basename.
	 */
	 
	result = g_string_new (NULL);
	g_string_append (result, priv->new_path);
	g_string_append (result, "/");
	g_string_append (result, new_basename);
	g_free (new_basename);
	
	return g_string_free (result, FALSE);
}

static void
op_finished (GPid pid, gint status, gpointer data)
{
	NfcConverter	    *converter;
	NfcConverterPrivate *priv;
	gboolean	     retry;

	converter = NFC_CONVERTER (data);
	priv = NFC_CONVERTER_GET_PRIVATE (converter);
	retry = TRUE;

	if (status != 0) {
		/* Conversion failed */
		NautilusFileInfo *file;
		gchar		 *name;
		GtkWidget	 *msg_dialog;
		int		  response_id;

		/* Get the name of the file that failed */
		file = NAUTILUS_FILE_INFO (priv->files->data);
		name = nautilus_file_info_get_name (file);

		/* Now create a message dialog to inform the user */
		/* TODO: Give the user a better message as to why */
		/*       the conversion failed.			  */
		msg_dialog =
		  gtk_message_dialog_new (GTK_WINDOW (priv->progress_dialog),
					  GTK_DIALOG_DESTROY_WITH_PARENT,
					  GTK_MESSAGE_ERROR,
					  GTK_BUTTONS_NONE,
					  "'%s' cannot be converted.",
					  name);
		g_free (name);

		gtk_dialog_add_button (GTK_DIALOG (msg_dialog), _("_Skip"), 1);
		gtk_dialog_add_button (GTK_DIALOG (msg_dialog), GTK_STOCK_CANCEL,
				       GTK_RESPONSE_CANCEL);
		gtk_dialog_add_button (GTK_DIALOG (msg_dialog), _("_Retry"), 0);
		gtk_dialog_set_default_response (GTK_DIALOG (msg_dialog), 0);

		response_id = gtk_dialog_run (GTK_DIALOG (msg_dialog));
		gtk_widget_destroy (msg_dialog);

		if (response_id == 0) {
			retry = TRUE;
		} else if (response_id == GTK_RESPONSE_CANCEL) {
			priv->cancelled = TRUE;
		} else if (response_id == 1) {
			retry = FALSE;
		}

	}

	if (status == 0 || !retry) {
		/* Flac has been successfully been converted (or skipped) */
		priv->flacs_converted++;
		priv->files = priv->files->next;
	}

	if (!priv->cancelled && priv->files != NULL) {
		/* Convert flac */
		run_op (converter);
	} else {
		/* Cancel|Teminate operation */
		gtk_widget_destroy (priv->progress_dialog);
	}
}

static void
nfc_converter_cancel_cb (GtkDialog *dialog,
			 gint       response_id,
			 gpointer   user_data)
{
	NfcConverter	    *converter;
	NfcConverterPrivate *priv;

	converter = NFC_CONVERTER (user_data);
	priv = NFC_CONVERTER_GET_PRIVATE (converter);

	priv->cancelled = TRUE;
	gtk_dialog_set_response_sensitive (dialog,
					   GTK_RESPONSE_CANCEL,
					   FALSE);
}

static void
create_progress_dialog (NfcConverter *converter)
{
	NfcConverterPrivate *priv;
	GtkWidget	    *vbox;
	GtkWidget	    *label;
	
	priv = NFC_CONVERTER_GET_PRIVATE (converter);

	priv->progress_dialog = gtk_dialog_new ();
	gtk_window_set_title (GTK_WINDOW (priv->progress_dialog),
			      _("Converting files"));
	gtk_dialog_add_button (GTK_DIALOG (priv->progress_dialog),
			       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL);
	gtk_dialog_set_has_separator (GTK_DIALOG (priv->progress_dialog),
				      FALSE);
	g_signal_connect (priv->progress_dialog, "response",
			  G_CALLBACK (nfc_converter_cancel_cb),
			  converter);

	vbox = GTK_DIALOG (priv->progress_dialog)->vbox;
	gtk_container_set_border_width (GTK_CONTAINER (priv->progress_dialog), 5);
	gtk_box_set_spacing (GTK_BOX (vbox), 8);
	gtk_window_set_default_size (GTK_WINDOW (priv->progress_dialog), 400, -1);

	label = gtk_label_new (_("<big><b>Converting files</b></big>"));
	gtk_label_set_use_markup (GTK_LABEL (label), TRUE);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0);
	gtk_box_pack_start_defaults (GTK_BOX (vbox), label);

	priv->progress_bar = gtk_progress_bar_new ();
	gtk_box_pack_start (GTK_BOX (vbox), priv->progress_bar, FALSE, FALSE, 0);

	priv->progress_label = gtk_label_new ("");
	gtk_misc_set_alignment (GTK_MISC (priv->progress_label), 0, 0);
	gtk_box_pack_start_defaults (GTK_BOX (vbox), priv->progress_label);

	gtk_widget_show_all (priv->progress_dialog);
}

static GtkWidget*
create_overwrite_dialog (GtkWidget *parent, gchar *filename)
{
	GtkWidget *dialog;
	GtkWidget *label;
	GtkWidget *vbox;
	gchar     *text;

	/* Create Dialog */
	dialog = gtk_dialog_new_with_buttons (_("Overwrite File"),
					      GTK_WINDOW (parent),
					      GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
					      GTK_STOCK_YES,
					      GTK_RESPONSE_YES,
					      GTK_STOCK_NO,
					      GTK_RESPONSE_NO,
					      NULL);

	/* Create the label */
	text = g_strdup_printf ("%s already exists.\nDo you wish to overwrite it?",
				filename);
	label = gtk_label_new (text);
	g_free (text);

	/* Set up vbox */
	vbox = GTK_DIALOG (dialog)->vbox;
	gtk_container_set_border_width (GTK_CONTAINER (dialog), 5);
	gtk_box_set_spacing (GTK_BOX (vbox), 8);

	/* Now add the label to the dialog */
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox),
			   label);
	gtk_widget_show (label);

	return dialog;
}

static void
run_op (NfcConverter *converter)
{
	NfcConverterPrivate *priv;

	priv = NFC_CONVERTER_GET_PRIVATE (converter);

	g_return_if_fail (priv->files != NULL);

	NautilusFileInfo *file_info;
	GFile		 *old_file;
	GFile		 *new_file;
	gchar 		 *uri;
	gchar		 *new_uri;
	gchar		 *old_filename;
	gchar		 *new_filename;
	gchar 		 *tmp;
	gchar		 *file_name;
	gchar		 *argv[7];
	pid_t		  pid;

	/* Create the progress dialog, if not present */
	if (priv->progress_dialog == NULL) {
		create_progress_dialog (converter);
	}

	file_info = NAUTILUS_FILE_INFO (priv->files->data);

	/* Create the GFile object for the original file */
	uri = nautilus_file_info_get_uri (file_info);
	old_file = g_file_new_for_uri (uri);
	g_free (uri);

	/* Let create the uri for the new file */	
	new_uri = nfc_converter_transform_uri (converter, old_file);
	
	/* Create the GFile object for the new file */
	new_file = g_file_new_for_uri (new_uri);
	g_free (new_uri);
	
	/* 
	 * Let's see if the new file already exists, and if 
	 * so ask the user whether they want to overwrite it
	 */
	if (g_file_query_exists (new_file, NULL)) {
		GtkWidget *dialog;
		gchar     *f;
		gint       result;

		f = g_file_get_basename (new_file);
		dialog = create_overwrite_dialog (priv->progress_dialog, f);
		g_free (f);

		result = gtk_dialog_run (GTK_DIALOG (dialog));

		switch (result) {
		case GTK_RESPONSE_NO:
			priv->flacs_converted++;
			priv->files = priv->files->next;

			g_object_unref (old_file);
			g_object_unref (new_file);
			gtk_widget_destroy (dialog);

			if (priv->files != NULL) {
				/* Convert next flac */
				run_op (converter);
			} else {
				/* Cancel|Teminate operation */
				gtk_widget_destroy (priv->progress_dialog);
			}

			return;
		default:
			gtk_widget_destroy (dialog);
		}
	}
	
	/* Get the local path for the original file */
	old_filename = g_file_get_path (old_file);
	g_object_unref (old_file);
	
	/* Get the local path for the new file */
	new_filename = g_file_get_path (new_file);
	g_object_unref (new_file);	

	/* Create the system call command */
	argv[0] = "/usr/bin/oggenc";
	argv[1] = "-q";
	argv[2] = priv->quality;
	argv[3] = old_filename;
	argv[4] = "-o";
	argv[5] = new_filename;
	argv[6] = NULL;

	/* Let's actually get to converting the files */
	if (!g_spawn_async (NULL, argv, NULL, G_SPAWN_DO_NOT_REAP_CHILD, NULL, NULL, &pid, NULL)) {
		/* FIXME: error handling */
		return;
	}
	
	g_free (old_filename);
	g_free (new_filename);

	g_child_watch_add (pid, op_finished, converter);

	/* TODO: Move the progress bar stuff to it's own function */
	/* Set the progress bar for the user */
	gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (priv->progress_bar),
				       (double) (priv->flacs_converted +1) / priv->flacs_total);
	tmp = g_strdup_printf (_("Converting FLAC: %d of %d"),
			       priv->flacs_converted +1,
			       priv->flacs_total);
	gtk_progress_bar_set_text (GTK_PROGRESS_BAR (priv->progress_bar), tmp);
	g_free (tmp);

	/* Let's set the name of the file we are currently converting */
	file_name = nautilus_file_info_get_name (file_info);
	tmp = g_strdup_printf (_("<i>Converting \"%s\"</i>"), file_name);
	g_free (file_name);
	gtk_label_set_markup (GTK_LABEL (priv->progress_label), tmp);
	g_free (tmp);
}

static void
nfc_converter_response_cb (GtkDialog *dialog,
			   gint       response_id,
			   gpointer   user_data)
{
	NfcConverter	    *converter;
	NfcConverterPrivate *priv;

	converter = NFC_CONVERTER (user_data);
	priv = NFC_CONVERTER_GET_PRIVATE (converter);

	if (response_id == GTK_RESPONSE_OK) {
		gint value;

		/* Get Ogg encoding quality */
		value =	gtk_spin_button_get_value_as_int (priv->quality_spinbutton);
		sprintf (priv->quality, "%i", value);

		/* Grab the save path */
		priv->new_path = g_strdup (gtk_file_chooser_get_uri (priv->path_chooser));

		/* Ok, let's get ready to rumble */
		run_op (converter);
	}
	gtk_widget_destroy (GTK_WIDGET (dialog));
}

static void
nfc_converter_init (NfcConverter *converter)
{
	NfcConverterPrivate *priv;
	GtkBuilder          *ui = NULL;
	GError              *err = NULL;
	gchar               *path;

	priv = NFC_CONVERTER_GET_PRIVATE (converter);

	/* Let's create our gtkbuilder and load the xml file */
	ui = gtk_builder_new ();
	path = g_build_filename (DATADIR, PACKAGE, "nfc.xml", NULL);
	gtk_builder_add_from_file (ui, path, &err);
	g_free (path);
	gtk_builder_set_translation_domain (ui, GETTEXT_PACKAGE);

	/* Grab some widgets */
	priv->dialog = GTK_DIALOG (gtk_builder_get_object (ui, "ogg_dialog"));
	priv->quality_spinbutton =
		GTK_SPIN_BUTTON (gtk_builder_get_object (ui,
							 "quality_spinbutton"));
	priv->path_chooser =
		GTK_FILE_CHOOSER (gtk_builder_get_object (ui,
							  "path_chooser"));

	/* Connect the siganl */	
	g_signal_connect (G_OBJECT (priv->dialog), "response",
			  (GCallback) nfc_converter_response_cb,
			  converter);
}


NfcConverter *
nfc_converter_new (GList *files)
{
	return g_object_new (NFC_TYPE_CONVERTER, "files", files, NULL);
}

void
nfc_converter_show_dialog (NfcConverter *converter)
{
	NfcConverterPrivate *priv;

	priv = NFC_CONVERTER_GET_PRIVATE (converter);
	gtk_widget_show (GTK_WIDGET (priv->dialog));
}
